/*
 * CAN.c
 *
 *  Created on: 18. 2. 2021
 *      Author: Ondra
 */

#include "CAN.h"
#include "stm32l4xx.h"
#include "Flash_Access.h"

extern CAN_HandleTypeDef hcan1;
extern CAN_FilterTypeDef sFilterConfig;
extern uint32_t TxMailbox;
extern CAN_TxHeaderTypeDef pTxHeader;
extern CAN_RxHeaderTypeDef pRxHeader;
extern uint8_t CAN_Rx_Buffer[8];
extern uint8_t CAN_Tx_Buffer[8];
extern ADC_HandleTypeDef hadc1;
extern ADC_HandleTypeDef hadc2;
extern uint16_t memChargeCurrent;
extern uint16_t memChargeVoltage;
extern uint16_t memInputCurrent;
extern uint16_t memChargeOption;
extern uint8_t memBatteryPeriod;
/*
 * posle zpravu po canu
 * vychozi std ID a data frame
 * vstupy: ID 11 bit, DLC delka dat, buffer pointer na data
 */
void CAN_Send_Message(uint16_t ID, uint8_t DLC, uint8_t *p_buffer) {
	pTxHeader.DLC = DLC;	//pocet bytu k prenosu (1-8 bytu)
	pTxHeader.IDE = CAN_ID_STD;	//volba ID, standard nebo extended
	pTxHeader.RTR = CAN_RTR_DATA;	//volba Data nebo Remote Frame
	pTxHeader.StdId = ID;	//IDcko, mezi 0x00 a 0x7FF => 11 bitu

	HAL_CAN_AddTxMessage(&hcan1, &pTxHeader, p_buffer, &TxMailbox); //sestavit message a odeslat
}
/*
 * posle remote frame po canu
 * vstupy: ID 11 bit, jen pro ladeni remote frame
 */
void CAN_Send_Remote(uint16_t ID) {
	uint8_t p_buffer;
	pTxHeader.IDE = CAN_ID_STD;	//volba ID, standard nebo extended
	pTxHeader.RTR = CAN_RTR_REMOTE;	//volba Data nebo Remote Frame
	pTxHeader.StdId = ID;	//IDcko, mezi 0x00 a 0x7FF => 11 bitu
	HAL_CAN_AddTxMessage(&hcan1, &pTxHeader, &p_buffer, &TxMailbox); //sestavit message a odeslat
}

/*Inicializace filtru CAN
 * Zde nastavit vsechny id pro smer z pocitace do jednotky, aby se zprava dostala pres filtry
 *
 */
void CAN_Filter_init(void) {
	//filtry se daj inicializovat i mimo INIT mod, ale zapnout FINIT bit
	//BANK 0
	sFilterConfig.FilterBank = 0; //cislo filterbank, 0-13
	sFilterConfig.FilterMode = CAN_FILTERMODE_IDLIST; //LIST mode
	sFilterConfig.FilterScale = CAN_FILTERSCALE_16BIT; //16 bit	=>strana 1632, uplne spodni pripad 4x16bit filtry
	sFilterConfig.FilterFIFOAssignment = CAN_FILTER_FIFO0; //fifo prirazene k filtru, bude to padat do FIFO 0
	sFilterConfig.FilterActivation = CAN_FILTER_ENABLE;
	//FMI - Filter Mask Index -> rika kterym filtrem to propadlo, cislo zalezi na rozlozeni LIST/MASK 16/32bit
	//Pokud je nektery filtr nepouzity a uvedu 0, propadne ID 0!!!
	sFilterConfig.FilterIdHigh = (ID_SMB_WW << 5);	//STID[10:3] STID[2:0] RTR IDE EXID[17:15], viz strana 1632, FMI2
	sFilterConfig.FilterIdLow = (ID_SMB_RW << 5); // FMI0
	sFilterConfig.FilterMaskIdHigh = (ID_SMB_WB_info << 5);  //FMI3
	sFilterConfig.FilterMaskIdLow = (ID_SMB_WB_data1 << 5);  //FMI1
	HAL_CAN_ConfigFilter(&hcan1, &sFilterConfig);	//zapsat

	//BANK1
	sFilterConfig.FilterBank = 1; //cislo filterbank, 0-13
	sFilterConfig.FilterMode = CAN_FILTERMODE_IDLIST; //LIST mode
	sFilterConfig.FilterScale = CAN_FILTERSCALE_16BIT; //16 bit	=>strana 1632, uplne spodni pripad 4x16bit filtry
	sFilterConfig.FilterFIFOAssignment = CAN_FILTER_FIFO0; //fifo prirazene k filtru, bude to padat do FIFO 0
	sFilterConfig.FilterActivation = CAN_FILTER_ENABLE;
	//FMI - Filter Mask Index -> rika kterym filtrem to propadlo, cislo zalezi na rozlozeni LIST/MASK 16/32bit
	//Pokud je nektery filtr nepouzity a uvedu 0, propadne ID 0!!!
	sFilterConfig.FilterIdHigh = (ID_SMB_WB_data2 << 5);	//STID[10:3] STID[2:0] RTR IDE EXID[17:15], viz strana 1632,
	sFilterConfig.FilterIdLow = (ID_SMB_WB_data3 << 5); //
	sFilterConfig.FilterMaskIdHigh = (ID_SMB_WB_data4 << 5);  //
	sFilterConfig.FilterMaskIdLow = (ID_SMB_WB_data5 << 5);  //
	HAL_CAN_ConfigFilter(&hcan1, &sFilterConfig);	//zapsat

	//BANK2
	sFilterConfig.FilterBank = 2; //cislo filterbank, 0-13
	sFilterConfig.FilterMode = CAN_FILTERMODE_IDLIST; //LIST mode
	sFilterConfig.FilterScale = CAN_FILTERSCALE_16BIT; //16 bit	=>strana 1632, uplne spodni pripad 4x16bit filtry
	sFilterConfig.FilterFIFOAssignment = CAN_FILTER_FIFO0; //fifo prirazene k filtru, bude to padat do FIFO 0
	sFilterConfig.FilterActivation = CAN_FILTER_ENABLE;
	//FMI - Filter Mask Index -> rika kterym filtrem to propadlo, cislo zalezi na rozlozeni LIST/MASK 16/32bit
	//Pokud je nektery filtr nepouzity a uvedu 0, propadne ID 0!!!
	sFilterConfig.FilterIdHigh = (ID_SMB_RB << 5);	//STID[10:3] STID[2:0] RTR IDE EXID[17:15], viz strana 1632,
	sFilterConfig.FilterIdLow = (ID_Settings << 5); //
	sFilterConfig.FilterMaskIdHigh = (ID_Analog_read << 5 | 0x10);  //POZOR, RTR frame
	sFilterConfig.FilterMaskIdLow = (ID_Battery_period << 5);  //
	HAL_CAN_ConfigFilter(&hcan1, &sFilterConfig);	//zapsat
}

/*
 * roztridi zpravu podle ID a zavola odpovidajici funkci
 */
void CAN_msg_sort(void) {
	HAL_CAN_GetRxMessage(&hcan1, CAN_RX_FIFO0, &pRxHeader, &CAN_Rx_Buffer[0]); //presunout message z Rx fifa do RAM
	switch (pRxHeader.StdId)	//rozhodovani podle ID
	{
	case ID_SMB_WW:	//write word
		CAN_WW_handle();
		break;
	case ID_SMB_RW:	//read word
		CAN_RW_handle();
		break;
	case ID_SMB_WB_info: //write block info - zacatek sekvence + write block data
		CAN_WB_handle();
		break;
	case ID_SMB_RB:	//read block, prijmu pozadavek, vyctu z SMBusu a vratim RB info + RB data
		CAN_RB_handle();
		break;
	case ID_Settings:	//nastaveni vychozich hodnot do flash
		CAN_Settings_handle();
		break;
	case ID_Analog_read:	//Analogove hodnoty odeslat do PC
		CAN_Analog_handle();
		break;
	case ID_Battery_period:	//Zmenit periodu odesilani dat
		CAN_Battery_period_handle();
		break;
	default:
		break;
	}
}

void CAN_WW_handle(void) {//obsluha protokolu writeword po canu, pokud nesedi dlc, tak error
	uint8_t CAN_msg = 0;
	if (pRxHeader.DLC == 4) {
		SMBus_Write_Word(CAN_Rx_Buffer[0], CAN_Rx_Buffer[1], ((CAN_Rx_Buffer[3] << 8) | CAN_Rx_Buffer[2]));
	} else {
		CAN_msg = Error_type_DLC;
		CAN_Send_Message(ID_Error, 1, &CAN_msg);
	}
}

void CAN_RW_handle(void) {//obsluha protokolu readword po canu, pokud nesedi dlc, tak error
	uint8_t CAN_msg = 0;
	uint16_t RWtemp = 0;
	if (pRxHeader.DLC == 2) {
		RWtemp = SMBus_Read_Word(CAN_Rx_Buffer[0], CAN_Rx_Buffer[1]);
		CAN_Tx_Buffer[0] = CAN_Rx_Buffer[0];	//adresa
		CAN_Tx_Buffer[1] = CAN_Rx_Buffer[1];	//command
		CAN_Tx_Buffer[2] = (uint8_t) RWtemp;	//low byte
		CAN_Tx_Buffer[3] = (uint8_t) (RWtemp >> 8);	//high byte
		CAN_Send_Message(ID_SMB_RW_reply, 4, CAN_Tx_Buffer);
	} else {
		CAN_msg = Error_type_DLC;
		CAN_Send_Message(ID_Error, 1, &CAN_msg);
	}
}

void CAN_WB_handle(void) {
	uint8_t CAN_msg = 0;
	uint8_t CAN_Rx_Buffer_long[40] = { 0 };
	uint8_t Addrtemp = 0;	//pro block
	uint8_t Commandtemp = 0;	//pro block
	uint8_t Lentemp = 0; //pro block
	uint8_t frames_r = 0;	//pocet prijatych can frames/nebo odeslanych
	uint8_t frames = 0;	//pro block
	uint32_t timeout = 0;
	uint8_t messageID = 0;	//cislo zpravy
	uint8_t framecheck = 0; //jednotlive bity se nahodi podle prijeti frame

	if (pRxHeader.DLC == 3) {
		Addrtemp = CAN_Rx_Buffer[0]; //adresa
		Commandtemp = CAN_Rx_Buffer[1]; //command
		Lentemp = CAN_Rx_Buffer[2]; //bytecount
		if (Lentemp > 1 && Lentemp <= 34) { //bytecount v rozmezi
			frames = Lentemp / 8;	//CAN_Rx_Buffer = pocet bajtu celkem - vypocitat pocet frame po canu, max 5 frames (34 bajtu)
			if (Lentemp % 8 != 0) {
				frames++;
			}

			timeout = HAL_GetTick() + 1000;	//1000ms timeout
			while (frames > frames_r && HAL_GetTick() < timeout) {	//chci prijmout vsechny frames, max 5
				if (HAL_CAN_GetRxFifoFillLevel(&hcan1, CAN_RX_FIFO0) > 0) {	//je neco na canu?
					HAL_CAN_GetRxMessage(&hcan1, CAN_RX_FIFO0, &pRxHeader, &CAN_Rx_Buffer[0]);
					messageID = (pRxHeader.StdId & 0x1F);	//id zprvy dle excelu, 14-18 obsahuje dat bytes
					if (messageID >= 14 && messageID <= 18) {	//spravne ID
						for (uint8_t i = 0; i < 8; i++) {
							CAN_Rx_Buffer_long[((messageID - 14) * 8) + i] = CAN_Rx_Buffer[i];	//presun do jednoho pole
						}
						framecheck = framecheck | (1 << (messageID - 14));
						frames_r++;
					}

					timeout = HAL_GetTick() + 1000;	//prodlouzeni timeoutu o 1000ms
				}

			}
			if (frames == frames_r) {	//prijal jsem vsechny frames, muzu to poslat po SMBusu
				if (framecheck == (0x1F >> (5 - frames))) {	//5 framu je maximum, sedi framecheck
					SMBus_Write_Block(Addrtemp, Commandtemp, &CAN_Rx_Buffer_long[0], Lentemp);	//zapisu block na SMBus
				}

			} else {
				//error - neprijal jsem vsechny frames
				CAN_msg = Error_type_Msg;
				CAN_Send_Message(ID_Error, 1, &CAN_msg);
			}
		} else {
			//error - spatny bytecount - CAN_Rx_Buffer[2]
		}
	} else {
		//error - spatna delka DLC
		CAN_msg = Error_type_DLC;
		CAN_Send_Message(ID_Error, 1, &CAN_msg);
	}

}

void CAN_RB_handle(void) {
	uint8_t CAN_msg = 0;
	uint8_t CAN_Tx_Buffer_long[40] = { 0 };
	uint8_t Lentemp = 0; //pro block
	uint8_t frames_r = 0;	//pocet prijatych can frames/nebo odeslanych
	uint8_t frames = 0;	//pro block
	uint32_t timeout = 0;
	uint8_t byteinframe = 0; //kolik bajtu narvu do framu

	if (pRxHeader.DLC == 2) {	//vycist z BQ
		//CAN_Rx_Buffer[0] = adresa, CAN_Rx_Buffer[1] = command
		Lentemp = SMBus_Read_Block(CAN_Rx_Buffer[0], CAN_Rx_Buffer[1], &CAN_Tx_Buffer_long[0]); //vraci block len, 0xFF je chyba,
		//data jdou do buffer_long, info jde do buffer
		CAN_Tx_Buffer[0] = CAN_Rx_Buffer[0];	//adresa
		CAN_Tx_Buffer[1] = CAN_Rx_Buffer[1];	//command
		CAN_Tx_Buffer[2] = Lentemp;	//bytecount vratim
		CAN_Send_Message(ID_SMB_RB_info, 3, &CAN_Tx_Buffer[0]);	//odesilam info RB, nasleduji data
		frames = Lentemp / 8;	//kolik potrebuju can_frames na odeslani blocku z smbusu
		if (Lentemp % 8 != 0) {
			frames++;
		}

		timeout = HAL_GetTick() + 1000;	//1000ms timeout
		while (frames > frames_r && HAL_GetTick() < timeout) {	//odesilam frames
			if (HAL_CAN_GetTxMailboxesFreeLevel(&hcan1) != 0) {	//jsou volne tx mailboxy, poslu
				if (frames_r == frames - 1) {	//plnim frame do maxima (8 bytu, posledni muze byt prazdnejsi, podle toho jak to vychazi)
					byteinframe = Lentemp - ((frames_r - 1) * 8);	//posledni frame
				} else {
					byteinframe = 8; //predchozi frames
				}
				CAN_Send_Message(ID_Read_base | (frames_r + 21), byteinframe, &CAN_Tx_Buffer_long[frames_r * 8]); //msg ID je 21 - 25
				frames_r++;
				timeout = HAL_GetTick() + 1000;	//prodlouzeni timeoutu o 1000ms
			}
			//HAL_CAN_GetTxMailboxesFreeLevel
		}
		if (frames_r == frames) {
			//povedlo se poslat vse
		} else {
			//nepovedlo se poslat vse, zahlcany can?
		}

	} else {
		//error, nesedi DLC
		CAN_msg = Error_type_DLC;
		CAN_Send_Message(ID_Error, 1, &CAN_msg);
	}
}

void CAN_Settings_handle(void) {
	uint8_t CAN_msg = 0;
	uint8_t memWrite[9] = {0};

	if (pRxHeader.DLC == 8) {
		for(int i = 0;i<8;i++){
			memWrite[i]= CAN_Tx_Buffer[i];	//priprava dat do pameti
		}
		memWrite[8] = memBatteryPeriod;
		save_to_flash(&memWrite[0], 9);	//zapsat do flashky
		//aktualizace hodnot - z flash do RAM
		memChargeCurrent = *(uint16_t*) (FLASH_STORAGE + 0); //prime sahnuti do pameti pres pointery, daji se brat i po vice bitech nez 8
		memChargeVoltage = *(uint16_t*) (FLASH_STORAGE + 2);
		memInputCurrent = *(uint16_t*) (FLASH_STORAGE + 4);
		memChargeOption = *(uint16_t*) (FLASH_STORAGE + 6);
	} else {
		CAN_msg = Error_type_DLC;
		CAN_Send_Message(ID_Error, 1, &CAN_msg);
	}
}

void CAN_Analog_handle(void) {	//vycteni hodnot VIN - napeti zdroje; IOUT chg a IOUT adp z BQ24725A
	//IOUT: (3300/4095*ADCvalI)/(0.01*20) [mA]
	//UIN:  (3300/4095*ADCvalU)*((294+36)/36) [mV]
	//ADC 12 bit -> 4095
	uint8_t CAN_msg = 0;
	uint16_t RW_temp = 0;

	uint16_t IOUTadp = 0;
	uint16_t IOUTchg = 0;
	uint16_t VIN = 0;
	uint16_t IOUTadp2 = 0;
	uint16_t IOUTchg2 = 0;
	uint16_t VIN2 = 0;
	uint8_t flip = 0;

	if (pRxHeader.RTR == CAN_RTR_REMOTE) {

		HAL_ADC_Start(&hadc1);	//zacni konverzi, Iout
		HAL_ADC_Start(&hadc2);	//zacni konverzi, VIN
		RW_temp = SMBus_Read_Word(BQ24725A_Address, BQ24725A_ChargeOption);	//vycist bity
		HAL_ADC_PollForConversion(&hadc1, 100);	// == HAL_OK) {	//pockej si na konec konverze
		if (RW_temp & 0x0020) {	//bit IOUT selection, 0= adapter current, 1=charge current
			IOUTchg = HAL_ADC_GetValue(&hadc1);	//charge current
			flip = 0;
		} else {
			IOUTadp = HAL_ADC_GetValue(&hadc1);	//adapter current
			flip = 1;
		}
		HAL_ADC_Stop(&hadc1);
		SMBus_Write_Word(BQ24725A_Address, BQ24725A_ChargeOption, RW_temp ^ 0x0020);	//negace bitu 5
		HAL_Delay(50); //pockat nez zmena zabere

		HAL_ADC_Start(&hadc1);	//zacni konverzi, Iout, druha hodnota
		HAL_ADC_PollForConversion(&hadc1, 100);	// == HAL_OK) {	//pockej si na konec konverze
		if (flip) {
			IOUTchg = HAL_ADC_GetValue(&hadc1);	//charge current
		} else {
			IOUTadp = HAL_ADC_GetValue(&hadc1);	//adapter current
		}
		HAL_ADC_Stop(&hadc1);

		SMBus_Write_Word(BQ24725A_Address, BQ24725A_ChargeOption, RW_temp);	//vratit puvodni data

		HAL_ADC_PollForConversion(&hadc2, 100);	// == HAL_OK) {	//pockej si na konec konverze
		VIN = HAL_ADC_GetValue(&hadc2);
		HAL_ADC_Stop(&hadc2);

		//UIN:  (3300/4095*ADCvalU)*((294+36)/36) [mV]
		//IOUT: (3300/4095*ADCvalI)/(0.01*20) [mA]
		//Uin[7:0]	Uin[15:8]	Iadp[7:0]	Iadp[15:8]	Ichg[7:0]	Ichg[15:8]

		VIN2 = (uint16_t)((3300.0 / 4095.0 * (float)VIN) * ((300.0 + 36.0) / 36.0));
		IOUTadp2 = (uint16_t)((3300.0 / 4095.0 * (float)IOUTadp) / (0.01 * 20.0));
		IOUTchg2 = (uint16_t)((3300.0 / 4095.0 * (float)IOUTchg) / (0.01 * 20.0));

		CAN_Tx_Buffer[0] = (uint8_t) (VIN2);
		CAN_Tx_Buffer[1] = (uint8_t) (VIN2 >> 8);
		CAN_Tx_Buffer[2] = (uint8_t) (IOUTadp2);
		CAN_Tx_Buffer[3] = (uint8_t) (IOUTadp2 >> 8);
		CAN_Tx_Buffer[4] = (uint8_t) (IOUTchg2);
		CAN_Tx_Buffer[5] = (uint8_t) (IOUTchg2 >> 8);
		CAN_Send_Message(ID_Analog, 6, &CAN_Tx_Buffer[0]);
	}

	else {	//neni rtr frame
		CAN_msg = Error_type_DLC;
		CAN_Send_Message(ID_Error, 1, &CAN_msg);
	}

}

void CAN_Battery_period_handle(void) {	//nastaveni periody pro odesilani zprav
	uint8_t CAN_msg = 0;
	uint8_t memWrite[9] = {0};

	if (pRxHeader.DLC == 1) {
		for(int i = 0;i<8;i++){
			memWrite[i]=*(uint8_t*) (FLASH_STORAGE + i);	//zaloha z pameti
		}
		memWrite[8] = CAN_Tx_Buffer[0];	//pridam to co prislo
		save_to_flash(&memWrite[0], 9);	//zapsat do flashky, zaloha + nova data
		memBatteryPeriod = *(uint8_t*) (FLASH_STORAGE + 8);

	} else {
		CAN_msg = Error_type_DLC;
		CAN_Send_Message(ID_Error, 1, &CAN_msg);
	}
}
